/*
 * Quaternion.xs
 *
 * Code here performs the basic contructor and math operations
 * on Quaternion objects.
 *
 * The C-code portion begins here.
 */
#include "EXTERN.h"
#include "perl.h"
#include "XSUB.h"
#include <math.h>

#define MEMLEAK_ID	46189

typedef struct {
	double wrh[4];
} QUATERNION;


#define QNORM2(x)	(x.wrh[0]*x.wrh[0]+x.wrh[1]*x.wrh[1]+x.wrh[2]*x.wrh[2]+x.wrh[3]*x.wrh[3])
#define QPNORM2(x)	(x->wrh[0]*x->wrh[0]+x->wrh[1]*x->wrh[1]+x->wrh[2]*x->wrh[2]+x->wrh[3]*x->wrh[3])


/*
 * int numeric_sv(SV *s, double *n)
 *
 * Determine if the SV is either a floating point or integer value, and
 * set that value in n.  Return 1 if it is a number, 0 if it is not.
 */
static int
numeric_sv(SV *s, double *n)
{
	*n = 0;

	if (SvTYPE(s) == SVt_NV)
		*n = SvNVX(s);
	else if (SvTYPE(s) == SVt_IV)
		*n = (double) SvIVX(s);
	else if (SvTYPE(s) == SVt_PV)
		*n = sv_2nv(s);
	else
		return 0;

	return 1;
}

/*
 * QUATERNION *qn_clone(QUATERNION *self)
 *
 * Create a copy of the quaternion struct.
 * Send out a warn() message and return NULL if there's no memory
 * to allocate.
 */
static QUATERNION *
qn_clone(QUATERNION *self)
{
	QUATERNION *cpy = NULL;

	if (self != NULL)
	{
		New(MEMLEAK_ID, cpy, 1, QUATERNION);	/* defined in handy.h */

		if (cpy)
		{
			cpy->wrh[0] = self->wrh[0];
			cpy->wrh[1] = self->wrh[1];
			cpy->wrh[2] = self->wrh[2];
			cpy->wrh[3] = self->wrh[3];
		}
	}

	if (cpy == NULL)
		warn("Unable to allocate Math::Quaternion object");

	return cpy;
}

static void
qn_free( QUATERNION * q)
{
	Safefree( q );
}

/*
 * QUATERNION * qn_conjugate(QUATERNION *a, QUATERNION *b)
 *
 * a = conjugate b
 * Returns a;
 */
static QUATERNION *
qn_conjugate(QUATERNION *a, QUATERNION * b)
{
	a->wrh[0] = b->wrh[0];
	a->wrh[1] = -b->wrh[1];
	a->wrh[2] = -b->wrh[2];
	a->wrh[3] = -b->wrh[3];

	return a;
}

/*
 * QUATERNION * qn_invert(QUATERNION *a, QUATERNION *b)
 *
 * a = invert b
 * Returns a, or NULL if the norm == 0.
 */
static QUATERNION *
qn_invert(QUATERNION *a, QUATERNION * b)
{
	double n = QPNORM2(b);

	if (n == 0.0)
		return NULL;

	a->wrh[0] = b->wrh[0] / n;
	a->wrh[1] = -b->wrh[1] / n;
	a->wrh[2] = -b->wrh[2] / n;
	a->wrh[3] = -b->wrh[3] / n;

	return a;
}

/*
 * QUATERNION * qn_multiply(QUATERNION *a, QUATERNION *b, QUATERNION *c)
 *
 * a = b * c.  In that order, since quaternions do not commute.
 */
static QUATERNION *
qn_multiply(QUATERNION *a, QUATERNION *b, QUATERNION *c)
{
	double a0, a1, a2, a3;

	a0 = b->wrh[0] * c->wrh[0]
		- b->wrh[1] * c->wrh[1]
		- b->wrh[2] * c->wrh[2]
		- b->wrh[3] * c->wrh[3];

	a1 = b->wrh[0] * c->wrh[1]
		+ b->wrh[1] * c->wrh[0]
		+ b->wrh[2] * c->wrh[3]
		- b->wrh[3] * c->wrh[2];

	a2 = b->wrh[0] * c->wrh[2]
		- b->wrh[1] * c->wrh[3]
		+ b->wrh[2] * c->wrh[0]
		+ b->wrh[3] * c->wrh[1];

	a3 = b->wrh[0] * c->wrh[3]
		+ b->wrh[1] * c->wrh[2]
		- b->wrh[2] * c->wrh[1]
		+ b->wrh[3] * c->wrh[0];

	a->wrh[0] = a0;
	a->wrh[1] = a1;
	a->wrh[2] = a2;
	a->wrh[3] = a3;

	return a;
}

/*
 * QUATERNION * qn_divide(QUATERNION *a, QUATERNION *b, QUATERNION *c)
 *
 * a = b / c.
 */
static QUATERNION *
qn_divide(QUATERNION *a, QUATERNION *b, QUATERNION *c)
{
	QUATERNION denominator;
#if 1
	double n = QPNORM2(c);

	if (n == 0.0)
		return NULL;

	qn_conjugate(&denominator, c);

	/*
	 * The denominator is now its own inverse * n.  Multiply it to the numerator.
	 */
	qn_multiply(a, b, &denominator);
	a->wrh[0] /= n;
	a->wrh[1] /= n;
	a->wrh[2] /= n;
	a->wrh[3] /= n;
	return a;
#else
	if (qn_invert(&denominator, c) == NULL)
		return NULL;

	/*
	 * The denominator is now its own inverse.  Multiply it to the numerator.
	 */
	return qn_multiply(a, b, &denominator);
#endif
}

/*
 * int qn_equal(QUATERNION *a, QUATERNION *b)
 *
 */
static int
qn_equal(QUATERNION *a, QUATERNION *b)
{
	return a->wrh[0] == b->wrh[0] &&
		a->wrh[1] == b->wrh[1] &&
		a->wrh[2] == b->wrh[2] &&
		a->wrh[3] == b->wrh[3];
}

/*
 * int qn_cmp(QUATERNION *a, QUATERNION *b)
 *
 * Return -1 if a < b, 0 if a == b, or 1 if a > b.
 */
static int
qn_cmp(QUATERNION *a, QUATERNION *b)
{
	register int x;
	for (x = 0; x < 4; x++)
	{
		if (a->wrh[x] < b->wrh[x])
			return -1;
		if (a->wrh[x] > b->wrh[x])
			return 1;
	}
	return 0;
}



/*
 * Now Exiting C-Code Land.
 *
 * Welcome to xsubpp.
 */
MODULE = Math::Quaternion	PACKAGE = Math::Quaternion	PREFIX = QTRN_
PROTOTYPES: ENABLE
REQUIRE: 1.95

#
# The constructor for this class.  Components default to zero if there
# are no values passed in.
#
QUATERNION *
QTRN_new(CLASS, ...)
	char *CLASS
    CODE:

	# defined in handy.h
	New(MEMLEAK_ID, RETVAL, 1, QUATERNION);

	if (RETVAL == NULL)
	{
		warn("Unable to allocate Math::Quaternion object");
		XSRETURN_UNDEF;
	}

	RETVAL->wrh[0] = RETVAL->wrh[1] = RETVAL->wrh[2] = RETVAL->wrh[3] = 0.0;

	switch (items)
	{
		case 5: RETVAL->wrh[3] = SvNV(ST(4));
		case 4: RETVAL->wrh[2] = SvNV(ST(3));
		case 3: RETVAL->wrh[1] = SvNV(ST(2));
		case 2: RETVAL->wrh[0] = SvNV(ST(1));
	}
    OUTPUT:
	RETVAL

#
# Construct a rotation quaternion by an angle around a vector.
#
QUATERNION *
QTRN_new_vrotate(CLASS, angle, b, c, d)
	char *CLASS
	double angle
	double b
	double c
	double d

    CODE:
    {
	double sin_angle = sin(angle/2.0);

	# defined in handy.h
	New(MEMLEAK_ID, RETVAL, 1, QUATERNION);

	if (RETVAL == NULL)
	{
		warn("Unable to allocate Math::Quaternion object");
		XSRETURN_UNDEF;
	}

	RETVAL->wrh[0] = cos(angle/2.0);
	RETVAL->wrh[1] = b * sin_angle;
	RETVAL->wrh[2] = c * sin_angle;
	RETVAL->wrh[3] = d * sin_angle;
    }

    OUTPUT:
	RETVAL

#
# ->_clone()
#
# Create a new copy of an existing quaternion object.
#
QUATERNION *
QTRN__clone(self)
	QUATERNION *self
    CODE:

	if ((RETVAL = qn_clone(self)) == NULL)
		XSRETURN_UNDEF;

    OUTPUT:
	RETVAL

#
# Supply a destructor for this class so that Perl knows
# how to destroy it.
#
void
DESTROY(self)
	QUATERNION *self
    CODE:
#	printf("# destroying %s\n", SvPV(ST(0),na) );
	qn_free( self );

#
# ->parts()
#
# Return the individual components of the quaternion as a list.
#
void
QTRN_parts(self)
	QUATERNION *self
    PPCODE:
	EXTEND(sp, 4);
	PUSHs(sv_2mortal(newSVnv(self->wrh[0])));
	PUSHs(sv_2mortal(newSVnv(self->wrh[1])));
	PUSHs(sv_2mortal(newSVnv(self->wrh[2])));
	PUSHs(sv_2mortal(newSVnv(self->wrh[3])));

#
# Perform '~self'.
#
QUATERNION *
QTRN__conjugate(self, nothing1, nothing2)
	QUATERNION *self
	SV *nothing1
	SV *nothing2
    CODE:
	qn_conjugate(self, self);
    OUTPUT:
	self

#
# Perform '-self'.
#
QUATERNION *
QTRN__negate(self, nothing1, nothing2)
	QUATERNION *self
	SV *nothing1
	SV *nothing2

    CODE:
	self->wrh[0] = -self->wrh[0];
	self->wrh[1] = -self->wrh[1];
	self->wrh[2] = -self->wrh[2];
	self->wrh[3] = -self->wrh[3];

    OUTPUT:
	self

#
# Perform abs(q).
#
SV *
QTRN_norm(self)
	QUATERNION *self
    CODE:
	RETVAL = newSVnv(sqrt(QPNORM2(self)));
    OUTPUT:
	RETVAL

#
# Perform  '$q1 = $q2 + $q3;',
#	or '$q1 += $q2;',
#	or '$q1++;'
#
QUATERNION *
QTRN__add(first, second, order)
	QUATERNION *first
	SV *second
	SV *order
    CODE:
    {
	double n;

	if ((RETVAL = qn_clone(first)) == NULL)
		XSRETURN_UNDEF;

	#
	# Handle the second argument by hand, as it may  be a
	# quaternion, or it may be just an ordinary number.
	# Ordinary numbers are treated as qtrn(number, 0, 0, 0).
	#
	if (sv_isobject(second) && (SvTYPE(SvRV(second)) == SVt_PVMG))
	{
		QUATERNION *other = (QUATERNION *)SvIV((SV*)SvRV( second ));
		RETVAL->wrh[0] += other->wrh[0];
		RETVAL->wrh[1] += other->wrh[1];
		RETVAL->wrh[2] += other->wrh[2];
		RETVAL->wrh[3] += other->wrh[3];
	}
	else if (numeric_sv(second, &n))
	{
		RETVAL->wrh[0] += n;
	}
	else
	{
		warn( "Operand not number or quaternion" );
		qn_free(RETVAL);
		XSRETURN_UNDEF;
	}
    }
    OUTPUT:
	RETVAL

#
# Perform  '$q1 = $q2 - $q3;',
#	or '$q1 -= $q2;',
#	or '$q1--;'
#
QUATERNION *
QTRN__sub(first, second, order)
	QUATERNION *first
	SV *second
	SV *order
    CODE:
    {
	double n;

	if ((RETVAL = qn_clone(first)) == NULL)
		XSRETURN_UNDEF;

	#
	# Handle the second argument by hand, as it may  be a
	# quaternion, or it may be just an ordinary number.
	# Ordinary numbers are treated as qtrn(number, 0, 0, 0).
	#
	if (sv_isobject(second) && (SvTYPE(SvRV(second)) == SVt_PVMG))
	{
		QUATERNION *other = (QUATERNION *)SvIV((SV*)SvRV( second ));
		RETVAL->wrh[0] -= other->wrh[0];
		RETVAL->wrh[1] -= other->wrh[1];
		RETVAL->wrh[2] -= other->wrh[2];
		RETVAL->wrh[3] -= other->wrh[3];
	}
	else if (numeric_sv(second, &n))
	{
		RETVAL->wrh[0] -= n;
	}
	else
	{
		warn( "Operand not number or quaternion" );
		qn_free(RETVAL);
		XSRETURN_UNDEF;
	}

	if (SvTRUE(order))
	{
		RETVAL->wrh[0] = -RETVAL->wrh[0];
		RETVAL->wrh[1] = -RETVAL->wrh[1];
		RETVAL->wrh[2] = -RETVAL->wrh[2];
		RETVAL->wrh[3] = -RETVAL->wrh[3];
	}
    }
    OUTPUT:
	RETVAL


#
# Perform  '$q1 = $q2 * $q3;',
#	or '$q1 *= $q2;',
#
QUATERNION *
QTRN__mul(first, second, order)
	QUATERNION *first
	SV *second
	SV *order

    CODE:
    {
	double n;

	if ((RETVAL = qn_clone(first)) == NULL)
		XSRETURN_UNDEF;

	#
	# Handle the second argument by hand, as it may  be a
	# quaternion, or it may be just an ordinary number.
	# Ordinary numbers are treated as qtrn(number, 0, 0, 0).
	#
	if (sv_isobject(second) && (SvTYPE(SvRV(second)) == SVt_PVMG))
	{
		QUATERNION *other = (QUATERNION *)SvIV((SV*)SvRV( second ));
		if (SvTRUE(order))
		{
			qn_multiply(RETVAL, other, first);
		}
		else
		{
			qn_multiply(RETVAL, first, other);
		}
	}
	else if (numeric_sv(second, &n))
	{
		#
		# Quaternion * number = Quaternion * qtrn(number, 0, 0, 0),
		# which works out to this:
		#
		RETVAL->wrh[0] *= n;
		RETVAL->wrh[1] *= n;
		RETVAL->wrh[2] *= n;
		RETVAL->wrh[3] *= n;
	}
	else
	{
		warn( "Operand not number or quaternion" );
		qn_free(RETVAL);
		XSRETURN_UNDEF;
	}
    }
    OUTPUT:
	RETVAL

#
# Perform  '$q1 = $q2 / $q3;',
#	or '$q1 /= $q2;',
#
QUATERNION *
QTRN__div(first, second, order)
	QUATERNION *first
	SV *second
	SV *order

    CODE:
    {
	QUATERNION numerator, denominator;
	double n;

	if ((RETVAL = qn_clone(first)) == NULL)
		XSRETURN_UNDEF;

	#
	# Handle the second argument by hand, as it may  be a
	# quaternion, or it may be just an ordinary number.
	#
	if (sv_isobject(second) && (SvTYPE(SvRV(second)) == SVt_PVMG))
	{
		if (SvTRUE(order))
			memcpy(&numerator, (QUATERNION *)SvIV((SV*)SvRV(second)), sizeof(QUATERNION));
		else
			memcpy(&denominator, (QUATERNION *)SvIV((SV*)SvRV(second)), sizeof(QUATERNION));
	}
	else if (numeric_sv(second, &n))
	{
		if (SvTRUE(order))
		{
			numerator.wrh[0] = n;
			numerator.wrh[1] = 0.0;
			numerator.wrh[2] = 0.0;
			numerator.wrh[3] = 0.0;
		}
		else
		{
			denominator.wrh[0] = n;
			denominator.wrh[1] = 0.0;
			denominator.wrh[2] = 0.0;
			denominator.wrh[3] = 0.0;
		}
	}

	if (SvTRUE(order))
		memcpy(&denominator, first, sizeof(QUATERNION));
	else
		memcpy(&numerator, first, sizeof(QUATERNION));


	if (qn_divide(RETVAL, &numerator, &denominator) == NULL)
	{
		warn("Illegal division by zero in Quaternion operation");
		qn_free(RETVAL);
		XSRETURN_UNDEF;
	}
    }

    OUTPUT:
	RETVAL

#
# Perform  '$q1 == $q2'
#
int
QTRN__equal(first, second, order)
	QUATERNION *first
	SV *second
	SV *order
    CODE:
    {
	double n;

	#
	# Handle the second argument by hand, as it may  be a
	# quaternion, or it may be just an ordinary number.
	# Ordinary numbers are treated as qtrn(number, 0, 0, 0).
	#
	if (sv_isobject(second) && (SvTYPE(SvRV(second)) == SVt_PVMG))
	{
		QUATERNION *other = (QUATERNION *)SvIV((SV*)SvRV( second ));
		RETVAL = qn_equal(first, other);
	}
	else if (numeric_sv(second, &n))
	{
		RETVAL = first->wrh[0] == n &&
			first->wrh[1] == 0.0 &&
			first->wrh[2] == 0.0 &&
			first->wrh[3] == 0.0;
	}
	else
	{
		warn( "Operand not number or quaternion" );
		XSRETURN_UNDEF;
	}
    }
    OUTPUT:
	RETVAL


#
# Perform  '$q1 != $q2'
#
int
QTRN__not_equal(first, second, order)
	QUATERNION *first
	SV *second
	SV *order
    CODE:
    {
	double n;

	#
	# Handle the second argument by hand, as it may  be a
	# quaternion, or it may be just an ordinary number.
	# Ordinary numbers are treated as qtrn(number, 0, 0, 0).
	#
	if (sv_isobject(second) && (SvTYPE(SvRV(second)) == SVt_PVMG))
	{
		QUATERNION *other = (QUATERNION *)SvIV((SV*)SvRV( second ));
		RETVAL = !qn_equal(first, other);
	}
	else if (numeric_sv(second, &n))
	{
		RETVAL = first->wrh[0] != n ||
			first->wrh[1] != 0.0 ||
			first->wrh[2] != 0.0 ||
			first->wrh[3] != 0.0;
	}
	else
	{
		warn( "Operand not number or quaternion" );
		XSRETURN_UNDEF;
	}
    }
    OUTPUT:
	RETVAL

#
# Perform  '$q1 < $q2'
#
int
QTRN__less_than(first, second, order)
	QUATERNION *first
	SV *second
	SV *order
    CODE:
    {
	QUATERNION q;
	double n;

	#
	# Handle the second argument by hand, as it may  be a
	# quaternion, or it may be just an ordinary number.
	# Ordinary numbers are treated as qtrn(number, 0, 0, 0).
	#
	if (sv_isobject(second) && (SvTYPE(SvRV(second)) == SVt_PVMG))
	{
		QUATERNION *other = (QUATERNION *)SvIV((SV*)SvRV( second ));
		RETVAL = qn_cmp(first, other) == -1;
	}
	else if (numeric_sv(second, &n))
	{
		q.wrh[0] = n;
		q.wrh[1] = q.wrh[2] = q.wrh[3] = 0.0;
		RETVAL = qn_cmp(first, &q) == -1;
	}
	else
	{
		warn( "Operand not number or quaternion" );
		XSRETURN_UNDEF;
	}
    }
    OUTPUT:
	RETVAL

#
# Perform  '$q1 <= $q2'
#
int
QTRN__less_equal(first, second, order)
	QUATERNION *first
	SV *second
	SV *order
    CODE:
    {
	QUATERNION q;
	double n;

	#
	# Handle the second argument by hand, as it may  be a
	# quaternion, or it may be just an ordinary number.
	# Ordinary numbers are treated as qtrn(number, 0, 0, 0).
	#
	if (sv_isobject(second) && (SvTYPE(SvRV(second)) == SVt_PVMG))
	{
		QUATERNION *other = (QUATERNION *)SvIV((SV*)SvRV( second ));
		RETVAL = qn_cmp(first, other) != 1;
	}
	else if (numeric_sv(second, &n))
	{
		q.wrh[0] = n;
		q.wrh[1] = q.wrh[2] = q.wrh[3] = 0.0;
		RETVAL = qn_cmp(first, &q) != 1;
	}
	else
	{
		warn( "Operand not number or quaternion" );
		XSRETURN_UNDEF;
	}
    }
    OUTPUT:
	RETVAL

#
# Perform  '$q1 > $q2'
#
int
QTRN__greater_than(first, second, order)
	QUATERNION *first
	SV *second
	SV *order
    CODE:
    {
	QUATERNION q;
	double n;

	#
	# Handle the second argument by hand, as it may  be a
	# quaternion, or it may be just an ordinary number.
	# Ordinary numbers are treated as qtrn(number, 0, 0, 0).
	#
	if (sv_isobject(second) && (SvTYPE(SvRV(second)) == SVt_PVMG))
	{
		QUATERNION *other = (QUATERNION *)SvIV((SV*)SvRV( second ));
		RETVAL = qn_cmp(first, other) == 1;
	}
	else if (numeric_sv(second, &n))
	{
		q.wrh[0] = n;
		q.wrh[1] = q.wrh[2] = q.wrh[3] = 0.0;
		RETVAL = qn_cmp(first, &q) == 1;
	}
	else
	{
		warn( "Operand not number or quaternion" );
		XSRETURN_UNDEF;
	}
    }
    OUTPUT:
	RETVAL

#
# Perform  '$q1 >= $q2'
#
int
QTRN__greater_equal(first, second, order)
	QUATERNION *first
	SV *second
	SV *order
    CODE:
    {
	QUATERNION q;
	double n;


	#
	# Handle the second argument by hand, as it may  be a
	# quaternion, or it may be just an ordinary number.
	# Ordinary numbers are treated as qtrn(number, 0, 0, 0).
	#
	if (sv_isobject(second) && (SvTYPE(SvRV(second)) == SVt_PVMG))
	{
		QUATERNION *other = (QUATERNION *)SvIV((SV*)SvRV( second ));
		RETVAL = qn_cmp(first, other) != -1;
	}
	else if (numeric_sv(second, &n))
	{
		q.wrh[0] = n;
		q.wrh[1] = q.wrh[2] = q.wrh[3] = 0.0;
		RETVAL = qn_cmp(first, &q) != -1;
	}
	else
	{
		warn( "Operand not number or quaternion" );
		XSRETURN_UNDEF;
	}
    }
    OUTPUT:
	RETVAL

#
# Perform  '$q1 <=> $q2'
#
int
QTRN__spaceship(first, second, order)
	QUATERNION *first
	SV *second
	SV *order
    CODE:
    {
	QUATERNION q;
	double n;

	#
	# Handle the second argument by hand, as it may  be a
	# quaternion, or it may be just an ordinary number.
	# Ordinary numbers are treated as qtrn(number, 0, 0, 0).
	#
	if (sv_isobject(second) && (SvTYPE(SvRV(second)) == SVt_PVMG))
	{
		QUATERNION *other = (QUATERNION *)SvIV((SV*)SvRV( second ));
		RETVAL = qn_cmp(first, other);
	}
	else if (numeric_sv(second, &n))
	{
		q.wrh[0] = n;
		q.wrh[1] = q.wrh[2] = q.wrh[3] = 0.0;
		RETVAL = qn_cmp(first, &q);
	}
	else
	{
		warn( "Operand not number or quaternion" );
		XSRETURN_UNDEF;
	}
    }
    OUTPUT:
	RETVAL

#
# Rotate a quaternion by a rotation described in another quaternion.
#
QUATERNION *
QTRN_rotate(self, qparam)
	QUATERNION *self
	SV *qparam

    CODE:
    {
	QUATERNION rq, inv_rq;

	if ((RETVAL = qn_clone(self)) == NULL)
		XSRETURN_UNDEF;

	if (sv_isobject(qparam) && (SvTYPE(SvRV(qparam)) == SVt_PVMG))
	{
		QUATERNION *other = (QUATERNION *)SvIV((SV*)SvRV( qparam ));
		rq.wrh[0] = other->wrh[0];
		rq.wrh[1] = other->wrh[1];
		rq.wrh[2] = other->wrh[2];
		rq.wrh[3] = other->wrh[3];
	}
	else
	{
		warn( "Parameter not a quaternion" );
		qn_free(RETVAL);
		XSRETURN_UNDEF;
	}

	if (qn_invert(&inv_rq, &rq) == NULL)
	{
		warn("Illegal division by zero in Quaternion operation");
		qn_free(RETVAL);
		XSRETURN_UNDEF;
	}

	qn_multiply(RETVAL, &inv_rq, RETVAL);
	qn_multiply(RETVAL, RETVAL, &rq);
    }

    OUTPUT:
	RETVAL
